@@ -100,7 +100,7 @@ class ThreespaceCommand:
                     response = response[size:]
                 else: #Strings are special, find the null terminator
                     str_len = response.index(0)
-                    output.append(struct.unpack(f"<{str_len}s", response[str_len])[0])
+                    output.append(struct.unpack(f"<{str_len}s", response[:str_len])[0])
                     response = response[str_len + 1:] #+1 to skip past the null terminator character too
         else: #Fast parse because no strings
             output.extend(struct.unpack(f"<{self.out_format}", response[:self.info.out_size]))
@@ -115,37 +115,42 @@ class ThreespaceCommand:
         raw = bytearray([])
         if self.info.num_out_params == 0: return None, raw
         output = []
-        for c in self.out_format:
-            if c != 's':
-                format_str = f"<{c}"
-                size = struct.calcsize(format_str)
-                response = com.read(size)
-                raw += response
-                if len(response) != size:
-                    if verbose:
-                        print(f"Failed to read {c} type. Aborting...")
-                    return None
-                output.append(struct.unpack(format_str, response)[0])
-            else: #Strings are special, find the null terminator
-                response = com.read(1)
-                raw += response
-                if len(response) != 1:
-                    if verbose:
-                        print(f"Failed to read string. Aborting...")
-                    return None
-                byte = chr(response[0])
-                string = ""
-                while byte != '\0':
-                    string += byte
-                    #Get next byte
-                    response = com.read(1)
+
+        if not math.isnan(self.info.out_size):
+            #Fast read and parse
+            response = com.read(self.info.out_size)
+            raw += response
+            if len(response) != self.info.out_size:
+                if verbose:
+                    print(f"Failed to read {self.info.name} {len(response)} / {self.info.out_size}. Aborting...")
+            output.extend(struct.unpack(f"<{self.out_format}", response))
+        else:
+            #There is a string, so go element by element
+            i = 0
+            while i < len(self.out_format):
+                c = self.out_format[i]
+                if c != 's':
+                    end_index = self.out_format.find('s', i)
+                    if end_index == -1: end_index = len(self.out_format)
+                    format_str = f"<{self.out_format[i:end_index]}"
+                    size = struct.calcsize(format_str)
+                    response = com.read(size)
                     raw += response
-                    if len(response) != 1:
+                    if len(response) != size:
+                        if verbose:
+                            print(f"Failed to read {c} type. Aborting...")
+                        return None
+                    output.append(struct.unpack(format_str, response)[0])
+                    i = end_index
+                else:
+                    response = com.read_until(b'\0')
+                    raw += response
+                    if response[-1] != 0:
                         if verbose:
                             print(f"Failed to read string. Aborting...")
                         return None
-                    byte = chr(response[0])
-                output.append(string)
+                    output.append(response[:-1].decode())
+                    i += 1
         
         if self.info.num_out_params == 1:
             return output[0], raw
@@ -162,7 +167,7 @@ class ThreespaceGetStreamingBatchCommand(ThreespaceCommand):
     def set_stream_slots(self, streaming_slots: list[ThreespaceCommand]):
         self.commands = streaming_slots
         self.out_format = ''.join(slot.out_format for slot in streaming_slots if slot is not None)
-        self.info.out_size = struct.calcsize(f"<{self.out_format}")
+        self.info.out_size = sum(slot.info.out_size for slot in streaming_slots if slot is not None)
 
     def parse_response(self, response: bytes):
         data = []
@@ -180,9 +185,8 @@ class ThreespaceGetStreamingBatchCommand(ThreespaceCommand):
         raw_response = bytearray([])
         for command in self.commands:
             if command is None: continue
-            binary = com.read(command.info.out_size)
-            raw_response += binary
-            out = command.parse_response(binary)
+            out, raw = command.read_command(com, verbose=verbose)
+            raw_response += raw
             response.append(out)
         
         return response, raw_response
@@ -393,6 +397,8 @@ class StreamableCommands(Enum):
     GetCorrectedAccelVec = 55
     GetCorrectedMagVec = 56
 
+    GetDateTime = 63
+
     GetRawGyroRate = 65
     GetRawAccelVec = 66
     GetRawMagVec = 67
@@ -401,8 +407,10 @@ class StreamableCommands(Enum):
     GetEeptsNewestStep = 71
     GetEeptsNumStepsAvailable = 72
 
+    GetDateTimeString = 93
     GetTimestamp = 94
 
+    GetBatteryCurrent = 200
     GetBatteryVoltage = 201
     GetBatteryPercent = 202
     GetBatteryStatus = 203
@@ -419,6 +427,10 @@ THREESPACE_AWAIT_COMMAND_FOUND = 0
 THREESPACE_AWAIT_COMMAND_TIMEOUT = 1
 THREESPACE_AWAIT_BOOTLOADER = 2
 
+THREESPACE_UPDATE_COMMAND_PARSED = 0
+THREESPACE_UPDATE_COMMAND_NOT_ENOUGH_DATA = 1
+THREESPACE_UPDATE_COMMAND_MISALIGNED = 2
+
 T = TypeVar('T')
 
 @dataclass
@@ -464,34 +476,43 @@ class ThreespaceBootloaderInfo:
 THREESPACE_REQUIRED_HEADER = THREESPACE_HEADER_ECHO_BIT | THREESPACE_HEADER_CHECKSUM_BIT | THREESPACE_HEADER_LENGTH_BIT
 class ThreespaceSensor:
     
-    def __init__(self, com = None, timeout=2, verbose=False):
+    def __init__(self, com = None, timeout=2, verbose=False, initial_clear_timeout=None):
         if com is None: #Default to attempting to use the serial com class if none is provided
             com = ThreespaceSerialComClass
-        
+        self.verbose = verbose
+
+        manually_opened_com = False
         #Auto discover using the supplied com class type
         if inspect.isclass(com) and issubclass(com, ThreespaceComClass):
             new_com = None
+            self.log("Auto-Discovering Sensor")
             for serial_com in com.auto_detect():
                 new_com = serial_com
                 break #Exit after getting 1
             if new_com is None:
-                raise RuntimeError("Failed to auto discover com port")
+                raise RuntimeError("Failed to auto discover com port")   
             self.com = new_com
+            manually_opened_com = True
             self.com.open()
         #The supplied com already was a com class, nothing to do
         elif inspect.isclass(type(com)) and issubclass(type(com), ThreespaceComClass):
             self.com = com
+            if not self.com.check_open():
+                self.com.open()
+                manually_opened_com = True
         else: #Unknown type, try making a ThreespaceSerialComClass out of this
             try:
                 self.com = ThreespaceSerialComClass(com)
             except:
                 raise ValueError("Failed to create default ThreespaceSerialComClass from parameter:", type(com), com)
 
+        self.restart_delay = 0.5
+
+        self.log("Configuring sensor communication")
         self.immediate_debug = True #Assume it is on from the start. May cause it to take slightly longer to initialize, but prevents breaking if it is on
         #Callback gives the debug message and sensor object that caused it
         self.__debug_cache: list[str] = [] #Used for storing startup debug messages until sensor state is confirmed
         
-        self.verbose = verbose
         self.debug_callback: Callable[[str, ThreespaceSensor],None] = self.__default_debug_callback
         self.misaligned = False
         self.dirty_cache = False
@@ -502,7 +523,15 @@ class ThreespaceSensor:
         self.is_data_streaming = False
         self.is_log_streaming = False
         self.is_file_streaming = False
+        self.log("Stopping potential streaming")
         self._force_stop_streaming()
+        #Clear out the buffer to allow faster initializing
+        #Ex: If a large buffer build up due to streaming, especially if using a slower interface like BLE,
+        #it may take a while before the entire garbage data can be parsed when checking for bootloader, causing a timeout
+        #even though it would have eventually succeeded   
+        self.log("Clearing com")     
+        self.__clear_com(initial_clear_timeout)
+
 
         #Used to ensure connecting to the correct sensor when reconnecting
         self.serial_number = None
@@ -514,12 +543,24 @@ class ThreespaceSensor:
         self.getStreamingBatchCommand: ThreespaceGetStreamingBatchCommand = None
         self.funcs = {}
 
-        self.__cached_in_bootloader = self.__check_bootloader_status()
-        if not self.in_bootloader:
-            self.__firmware_init()
-        else:
-            self.__cache_serial_number(self.bootloader_get_sn())
-            self.__empty_debug_cache()
+        self.log("Checking firmware status")
+        try:
+            self.__cached_in_bootloader = self.__check_bootloader_status()
+            if not self.in_bootloader:
+                self.log("Initializing firmware")
+                self.__firmware_init()
+            else:
+                self.log("Initializing bootloader")
+                self.__cache_serial_number(self.bootloader_get_sn())
+                self.__empty_debug_cache()
+        #This is just to prevent a situation where instantiating the API creates and fails to release a com class on failure when user catches the exception
+        #If user provides the com class, it is up to them to handle its state on error
+        except Exception as e:
+            self.log("Failed to initialize sensor")
+            if manually_opened_com:
+                self.com.close()
+            raise e
+        self.log("Successfully initialized sensor")
 
     #Just a helper for outputting information
     def log(self, *args):
@@ -528,6 +569,17 @@ class ThreespaceSensor:
 
 #-----------------------INITIALIZIATION & REINITIALIZATION-----------------------------------
 
+    def __clear_com(self, refresh_timeout=None):
+        data = self.com.read_all()
+        if refresh_timeout is None: return
+        while len(data) > 0: #Continue until all data is cleared
+            self.log(f"Refresh clear Length: {len(data)}")
+            start_time = time.time()
+            while time.time() - start_time < refresh_timeout: #Wait up to refresh time for a new message
+                data = self.com.read_all()
+                if len(data) > 0:
+                    break #Refresh the start time and wait for more data
+
     def __firmware_init(self):
         """
         Should only be called when not streaming and known in firmware.
@@ -565,14 +617,14 @@ class ThreespaceSensor:
         self.streaming_packet_size = 0
         self._force_stop_streaming()
 
-        #Now reinitialize the cached settings
-        self.__cache_header_settings()
-        self.__cache_streaming_settings()
-
         self.__cache_serial_number(int(self.get_settings("serial_number"), 16))
         self.__empty_debug_cache()
         self.immediate_debug = int(self.get_settings("debug_mode")) == 1 #Needed for some startup processes when restarting
 
+        #Now reinitialize the cached settings
+        self.__cache_header_settings()
+        self.__cache_streaming_settings()
+
     def __initialize_commands(self):
         self.commands: list[ThreespaceCommand] = [None] * 256
         self.getStreamingBatchCommand: ThreespaceGetStreamingBatchCommand = None
@@ -619,13 +671,16 @@ class ThreespaceSensor:
             method = types.MethodType(command.custom_func, self)
         else:
             #Build the actual method for executing the command
-            code = f"def {command.info.name}(self, *args):\n"
-            code += f"    return self.execute_command(self.commands[{command.info.num}], *args)"
+            code = f"def {command.info.name}(self, *args, **kwargs):\n"
+            code += f"    return self.execute_command(self.commands[{command.info.num}], *args, **kwargs)"
             exec(code, globals(), self.funcs)
             method = types.MethodType(self.funcs[command.info.name], self)
 
         setattr(self, command.info.name, method)
 
+    def has_command(self, command: ThreespaceCommand):
+        return self.commands[command.info.num] is not None 
+
     def __get_command(self, command_name: str):
         for command in self.commands:
             if command is None: continue
@@ -734,6 +789,17 @@ class ThreespaceSensor:
 
 #-----------------------------------------------BASE SETTINGS PROTOCOL------------------------------------------------
 
+    #Helper for converting python types to strings that set_settings can understand
+    def __internal_str(self, value):
+        if isinstance(value, float):
+            return f"{value:.10f}"
+        elif isinstance(value, bool):
+            return int(value)
+        elif isinstance(value, Enum):
+            return str(value.value)
+        else:
+            return str(value)        
+
     #Can't just do if "header" in string because log_header_enabled exists and doesn't actually require cacheing the header
     HEADER_KEYS = ["header", "header_status", "header_timestamp", "header_echo", "header_checksum", "header_serial", "header_length"]
     def set_settings(self, param_string: str = None, **kwargs):
@@ -745,10 +811,10 @@ class ThreespaceSensor:
         
         for key, value in kwargs.items():
             if isinstance(value, list):
-                value = [str(v) for v in value]
+                value = [self.__internal_str(v) for v in value]
                 value = ','.join(value)
-            elif isinstance(value, bool):
-                value = int(value)
+            else:
+                value = self.__internal_str(value)
             params.append(f"{key}={value}")
         cmd = f"!{';'.join(params)}\n"
 
@@ -757,8 +823,7 @@ class ThreespaceSensor:
             return 0xFF, 0xFF
 
         #For dirty check
-        keys = cmd[1:-1].split(';')
-        keys = [v.split('=')[0] for v in keys]
+        param_dict = threespaceSetSettingsStringToDict(cmd[1:-1])
 
         #Must enable this before sending the set so can properly handle reading the response
         if "debug_mode=1" in cmd:
@@ -786,21 +851,30 @@ class ThreespaceSensor:
         #Handle updating state variables based on settings
         #If the user modified the header, need to cache the settings so the API knows how to interpret responses
         if "header" in cmd.lower(): #First do a quick check
-            if any(v in keys for v in ThreespaceSensor.HEADER_KEYS): #Then do a longer check
+            if any(v in param_dict.keys() for v in ThreespaceSensor.HEADER_KEYS): #Then do a longer check
                 self.__cache_header_settings()
         
         if "stream_slots" in cmd.lower():
             self.__cache_streaming_settings()
         
         #All the settings changed, just need to mark dirty
-        if any(v in keys for v in ("default", "reboot")):
+        if any(v in param_dict.keys() for v in ("default", "reboot")):
             self.set_cached_settings_dirty()
 
         if err:
             self.log(f"Err setting {cmd}: {err=} {num_successes=}")
         return err, num_successes
 
-    def get_settings(self, *args: str) -> dict[str, str] | str:
+    def get_settings(self, *args: str, format="Mixed") -> dict[str, str] | str:
+        """
+        Gets the values for all requested settings. Settings are request by their string name. The result will be
+        the string response to that setting.
+
+        Params
+        -----
+        *args : Any number of string keys
+        format : "Mixed" (Dictionary if multiple settings requested, else just the response string) or "Dict" (Always a dictionary even if only one key)
+        """
         self.check_dirty()
         #Build and send the cmd
         params = list(args)
@@ -837,7 +911,7 @@ class ThreespaceSensor:
                 self.log("Failed to parse get value:", i, v, len(v))
         
         #Format response
-        if len(response_dict) == 1:
+        if len(response_dict) == 1 and format == "Mixed":
             return list(response_dict.values())[0]
         return response_dict
 
@@ -944,6 +1018,7 @@ class ThreespaceSensor:
         header = ThreespaceHeader.from_bytes(header, self.header_info)
         return header
 
+    #TODO: max_data_length is not sufficient. Need MINIMUM length as well. Can have a situation where len = 0 and checksum = 0
     def __peek_checksum(self, header: ThreespaceHeader, max_data_length=4096):
         """
         Using a header that contains the checksum and data length, calculate the checksum of the expected
@@ -956,11 +1031,18 @@ class ThreespaceSensor:
         """
         header_len = len(header.raw_binary)
         if header.length > max_data_length:
-            self.log("DATA TOO BIG:", header.length)
+            if not self.misaligned:
+                self.log("DATA TOO BIG:", header.length)
             return False
         data = self.com.peek(header_len + header.length)[header_len:]
-        if len(data) != header.length: return False
+        if len(data) != header.length: 
+            if not self.misaligned:
+                self.log(f"Data Length Mismatch - Got: {len(data)} Expected: {header.length}")
+            return False
         checksum = sum(data) % 256
+        if checksum != header.checksum and not self.misaligned:
+            self.log(f"Checksum Mismatch - Got: {checksum} Expected: {header.checksum}")
+            self.log(f"Data: {data}")
         return checksum == header.checksum
 
     def __await_command(self, cmd: ThreespaceCommand, timeout=2):
@@ -997,7 +1079,7 @@ class ThreespaceSensor:
 
 #------------------------------BASE INPUT PARSING--------------------------------------------
 
-    def __internal_update(self, header: ThreespaceHeader = None):
+    def __internal_update(self, header: ThreespaceHeader = None, blocking=True):
         """
         Manages checking the datastream for asynchronous responses (Streaming, Immediate Debug Messages).
         If no data is found to match these responses, the data buffer will be considered corrupted/misaligned
@@ -1012,29 +1094,39 @@ class ThreespaceSensor:
 
         Returns
         --------
-        False : Misalignment
-        True : Internal Data Found/Parsed
+        0 : Internal Data Found/Parsed
+        1 : Not enough data (Only possible when blocking == False)
+        2 : Misalignment
         """
         checksum_match = False #Just for debugging
 
         if header is not None:
             #NOTE: FOR THIS TO WORK IT IS REQUIRED THAT THE HEADER DOES NOT CHANGE WHILE STREAMING ANY FORM OF DATA.
             #IT IS UP TO THE API TO ENFORCE NOT ALLOWING HEADER CHANGES WHILE ANY OF THOSE THINGS ARE HAPPENING
-            if self.is_data_streaming and header.echo == THREESPACE_GET_STREAMING_BATCH_COMMAND_NUM: #Its a streaming packet, so update streaming
+            if self.is_data_streaming and header.echo == THREESPACE_GET_STREAMING_BATCH_COMMAND_NUM:
+                if not blocking:
+                    expected_output_size = len(header.raw_binary) + self.getStreamingBatchCommand.info.out_size
+                    if self.com.length < expected_output_size: return THREESPACE_UPDATE_COMMAND_NOT_ENOUGH_DATA
                 if checksum_match := self.__peek_checksum(header, max_data_length=self.getStreamingBatchCommand.info.out_size):
                     self.__update_base_streaming()
                     self.misaligned = False
-                    return True
+                    return THREESPACE_UPDATE_COMMAND_PARSED
             elif self.is_log_streaming and header.echo == THREESPACE_FILE_READ_BYTES_COMMAND_NUM:
-                if checksum_match := self.__peek_checksum(header, max_data_length=2560): #TODO: Confirm this number can be 2048 and then pound define it. Currently set to 2560 because I know that is big enough, but not optimal
+                if not blocking:
+                    expected_output_size = len(header.raw_binary) + min(header.length, THREESPACE_LIVE_LOG_STREAM_MAX_PACKET_SIZE)
+                    if self.com.length < expected_output_size: return THREESPACE_UPDATE_COMMAND_NOT_ENOUGH_DATA
+                if checksum_match := self.__peek_checksum(header, max_data_length=THREESPACE_LIVE_LOG_STREAM_MAX_PACKET_SIZE):
                     self.__update_log_streaming()
                     self.misaligned = False
-                    return True
+                    return THREESPACE_UPDATE_COMMAND_PARSED
             elif self.is_file_streaming and header.echo == THREESPACE_FILE_READ_BYTES_COMMAND_NUM:
+                if not blocking:
+                    expected_output_size = len(header.raw_binary) + min(header.length, THREESPACE_FILE_STREAMING_MAX_PACKET_SIZE)
+                    if self.com.length < expected_output_size: return THREESPACE_UPDATE_COMMAND_NOT_ENOUGH_DATA
                 if checksum_match := self.__peek_checksum(header, max_data_length=THREESPACE_FILE_STREAMING_MAX_PACKET_SIZE):
                     self.__update_file_streaming()
                     self.misaligned = False
-                    return True
+                    return THREESPACE_UPDATE_COMMAND_PARSED
         
         #Debug messages are possible and there is enough data to potentially be a debug message
         #NOTE: Firmware should avoid putting more then one \r\n in a debug message as they will be treated as unprocessed/misaligned characters
@@ -1049,7 +1141,7 @@ class ThreespaceSensor:
                     message = self.com.readline() #Read out the whole message!
                     self.debug_callback(message.decode('ascii'), self)
                     self.misaligned = False
-                    return True
+                    return THREESPACE_UPDATE_COMMAND_PARSED
 
         #The response didn't match any of the expected asynchronous streaming API responses, so assume a misalignment
         if header is not None:
@@ -1060,7 +1152,7 @@ class ThreespaceSensor:
             msg = "Possible Misalignment or corruption/debug message"
         #self.log("Misaligned:", self.com.peek(1))
         self.__handle_misalignment(msg)
-        return False
+        return THREESPACE_UPDATE_COMMAND_MISALIGNED
 
     def __handle_misalignment(self, message: str = None):
         if not self.misaligned and message is not None:
@@ -1124,9 +1216,9 @@ class ThreespaceSensor:
 
     def startStreaming(self) -> ThreespaceCmdResult[None]: ...
     def __startStreaming(self) -> ThreespaceCmdResult[None]:
-        if self.is_data_streaming: return
-        self.streaming_packets.clear()
-        self.__cache_streaming_settings()
+        if not self.is_data_streaming:
+            self.streaming_packets.clear()
+            self.__cache_streaming_settings()
 
         result = self.execute_command(self.commands[THREESPACE_START_STREAMING_COMMAND_NUM])
         self.is_data_streaming = True
@@ -1158,12 +1250,21 @@ class ThreespaceSensor:
         self.streaming_packets.clear()
     
     #This is called for all streaming types
-    def updateStreaming(self, max_checks=float('inf')):
+    def updateStreaming(self, max_checks=float('inf'), timeout=None, blocking=False):
         """
         Returns true if any amount of data was processed whether valid or not. This is called for all streaming types.
+
+        Parameters
+        ----------
+        max_checks : Will only attempt to read up to max_checks packets
+        timeout : Will only attempt to read packets for this duration. It is possible for this function to take longer then this timeout \
+        if blocking = True, in which case it could take up to timeout + com.timeout 
+        blocking : If False, will immediately stop when not enough data is available. If true, will immediately stop if not enough data \
+        for a header, but will block when trying to retrieve the data associated with that header. For most com classes, this does not matter. \
+        But for communication such as BLE where the header and data may be split between different packets, this will have a clear effect.
         """
         if not self.is_streaming: return False
-
+        if timeout is None: timeout = float('inf')
         #I may need to make this have a max num bytes it will process before exiting to prevent locking up on slower machines
         #due to streaming faster then the program runs
         num_checks = 0
@@ -1173,12 +1274,17 @@ class ThreespaceSensor:
                 return data_processed
             
             #Get header
+
             header = self.com.peek(self.header_info.size)
 
             #Get the header and send it to the internal update
             header = ThreespaceHeader.from_bytes(header, self.header_info)
-            self.__internal_update(header)
-            data_processed = True #Internal update always processes data. Either reads a streaming message, or advances buffer due to misalignment
+            result = self.__internal_update(header, blocking=blocking)
+            if result == THREESPACE_UPDATE_COMMAND_PARSED:
+                data_processed = True
+            elif result == THREESPACE_UPDATE_COMMAND_NOT_ENOUGH_DATA:
+                return data_processed
+            
             num_checks += 1
         
         return data_processed
@@ -1300,7 +1406,7 @@ class ThreespaceSensor:
         cmd.send_command(self.com)
         self.com.close()
         #TODO: Make this actually wait instead of an arbitrary sleep length
-        time.sleep(0.5) #Give it time to restart
+        time.sleep(self.restart_delay) #Give it time to restart
         self.com.open()
         self.__firmware_init()
 
@@ -1311,7 +1417,7 @@ class ThreespaceSensor:
         cmd = self.commands[THREESPACE_ENTER_BOOTLOADER_COMMAND_NUM]
         cmd.send_command(self.com)
         #TODO: Make this actually wait instead of an arbitrary sleep length
-        time.sleep(0.5) #Give it time to boot into bootloader
+        time.sleep(self.restart_delay) #Give it time to boot into bootloader
         if self.com.reenumerates:
             self.com.close()
             success = self.__attempt_rediscover_self()
@@ -1375,7 +1481,7 @@ class ThreespaceSensor:
     def bootloader_boot_firmware(self):
         if not self.in_bootloader: return
         self.com.write("B".encode())
-        time.sleep(0.5) #Give time to boot into firmware
+        time.sleep(self.restart_delay) #Give time to boot into firmware
         if self.com.reenumerates:
             self.com.close()
             success = self.__attempt_rediscover_self()
@@ -1392,13 +1498,14 @@ class ThreespaceSensor:
         This may take a long time
         """
         self.com.write('S'.encode())
-        if timeout is not None:
-            cached_timeout = self.com.timeout
-            self.com.timeout = timeout
-        response = self.com.read(1)[0]
-        if timeout is not None:
-            self.com.timeout = cached_timeout
-        return response
+
+        start_time = time.perf_counter()
+        response = []
+        while len(response) == 0 and time.perf_counter() - start_time < timeout:
+            response = self.com.read(1)
+        if len(response) == 0:
+            return -1
+        return response[0]
     
     def bootloader_get_info(self):
         self.com.write('I'.encode())
@@ -1408,14 +1515,20 @@ class ThreespaceSensor:
         bootversion = struct.unpack(f">{_3space_format_to_external('I')}", self.com.read(2))[0]
         return ThreespaceBootloaderInfo(memstart, memend, pagesize, bootversion)
 
-    def bootloader_prog_mem(self, bytes: bytearray):
+    def bootloader_prog_mem(self, bytes: bytearray, timeout=5):
         memsize = len(bytes)
         checksum = sum(bytes)
         self.com.write('C'.encode())
         self.com.write(struct.pack(f">{_3space_format_to_external('I')}", memsize))
         self.com.write(bytes)
         self.com.write(struct.pack(f">{_3space_format_to_external('B')}", checksum & 0xFFFF))
-        return self.com.read(1)[0]
+        start_time = time.perf_counter()
+        result = []
+        while len(result) == 0 and time.perf_counter() - start_time < timeout:
+            result = self.com.read(1)
+        if len(result) > 0:
+            return result[0]
+        return -1
 
     def bootloader_get_state(self):
         self.com.write('OO'.encode()) #O is sent twice to compensate for a bug in some versions of the bootloader where the next character is ignored (except for R, do NOT send R after O, it will erase all settings)
@@ -1427,20 +1540,26 @@ class ThreespaceSensor:
         self.com.write("RR".encode())
 
     def cleanup(self):
-        if not self.in_bootloader:
-            if self.is_data_streaming:
-                self.stopStreaming()
-            if self.is_file_streaming:
-                self.fileStopStream()
-            if self.is_log_streaming:
-                self.stopDataLogging()
+        error = None
+        try:
+            if not self.in_bootloader:
+                if self.is_data_streaming:
+                    self.stopStreaming()
+                if self.is_file_streaming:
+                    self.fileStopStream()
+                if self.is_log_streaming:
+                    self.stopDataLogging()
 
-            #The sensor may or may not have this command registered. So just try it
-            try:
-                #May not be opened, but also not cacheing that so just attempt to close.
-                self.closeFile()
-            except: pass
-        self.com.close()
+                #The sensor may or may not have this command registered. So just try it
+                try:
+                    #May not be opened, but also not cacheing that so just attempt to close.
+                    self.closeFile()
+                except: pass
+        except Exception as e:
+            error = e
+        self.com.close() #Ensuring the close gets called, that way com ports can't get stuck open. Also makes calling cleanup() "safe" even after disconnect
+        if error:
+            raise error
 
 #-------------------------START ALL PROTOTYPES------------------------------------
 
@@ -1525,6 +1644,7 @@ class ThreespaceSensor:
     def getCorrectedMagVec(self, id: int) -> ThreespaceCmdResult[list[float]]: ...    
     def enableMSC(self) -> ThreespaceCmdResult[None]: ...
     def disableMSC(self) -> ThreespaceCmdResult[None]: ...
+    def getDateTimeString(self) -> ThreespaceCmdResult[str]: ...
     def getTimestamp(self) -> ThreespaceCmdResult[int]: ...
     def getBatteryVoltage(self) -> ThreespaceCmdResult[float]: ...
     def getBatteryPercent(self) -> ThreespaceCmdResult[int]: ...
@@ -1653,6 +1773,7 @@ _threespace_commands: list[ThreespaceCommand] = [
     ThreespaceCommand("stopStreaming", THREESPACE_STOP_STREAMING_COMMAND_NUM, "", "", custom_func=ThreespaceSensor._ThreespaceSensor__stopStreaming),
     ThreespaceCommand("pauseLogStreaming", 87, "b", ""),
     
+    ThreespaceCommand("getDateTimeString", 93, "", "S"),
     ThreespaceCommand("getTimestamp", 94, "", "U"),
 
     ThreespaceCommand("tareWithCurrentOrientation", 96, "", ""),
@@ -1681,6 +1802,7 @@ _threespace_commands: list[ThreespaceCommand] = [
     ThreespaceCommand("fileStartStream", 180, "", "U", custom_func=ThreespaceSensor._ThreespaceSensor__fileStartStream),
     ThreespaceCommand("fileStopStream", 181, "", "", custom_func=ThreespaceSensor._ThreespaceSensor__fileStopStream),
 
+    ThreespaceCommand("getBatteryCurrent", 200, "", "I"),
     ThreespaceCommand("getBatteryVoltage", 201, "", "f"),
     ThreespaceCommand("getBatteryPercent", 202, "", "b"),
     ThreespaceCommand("getBatteryStatus", 203, "", "b"),
@@ -1729,4 +1851,17 @@ def threespaceGetHeaderLabels(header_info: ThreespaceHeaderInfo):
         order.append("serial#")
     if header_info.length_enabled:
         order.append("len")
-    return order
\ No newline at end of file
+    return order
+
+def threespaceSetSettingsStringToDict(setting_string: str):
+    d = {}
+    for item in setting_string.split(';'):
+        result = item.split('=')
+        key = result[0]
+        if len(result) == 1:
+            value = None
+        else:
+            value = '='.join(result[1:]) #In case = was part of the value, do a join
+        
+        d[key] = value
+    return d
\ No newline at end of file
